from .base import Application, Handler, Resource, ResourceConnector
from .connectors import SqsPermission, sqs
from .handlers import (
    JobEvent,
    QueueEvent,
    Request,
    Response,
    SqsEventsSequence,
    SqsFailedEvents,
    job,
    route,
    sqs_handler,
)
from .resources import Sqs



import json
from collections import UserList
from typing import Any, Callable, Iterable, Optional

from viburnum.application.base import Handler, LambdaInput, LambdaOutput
from viburnum.application.types import HeadersType, JsonData, MultiQueryParamsType

# ___________________ API __________________________
# Lambda with Rest Api
# https://docs.aws.amazon.com/lambda/latest/dg/services-apigateway.html#apigateway-example-event


class Request(LambdaInput):
    def __init__(self, event: dict, context: dict) -> None:
        super().__init__(event, context)
        self._json = None

    @property
    def raw_headers(self) -> HeadersType:
        return self.event["headers"]

    @property
    def raw_query_params(self) -> MultiQueryParamsType:
        return self.event["multiValueQueryStringParameters"]

    @property
    def method(self) -> str:
        return self.event["httpMethod"]

    @property
    def path(self) -> str:
        return self.event["path"]

    @property
    def path_params(self) -> dict:
        return self.event["pathParameters"] or {}

    @property
    def body(self) -> Optional[str]:
        return self.event["body"]

    def json(self) -> JsonData:
        if not self.body:
            return None
        if not self._json:
            self._json = json.loads(self.body)
        return self._json


class Response(LambdaOutput):
    def __init__(
        self,
        status_code: int,
        body: dict,
        headers: dict = None,
    ) -> None:
        # TODO: add support for multivalue headers
        self.response_data = {
            "statusCode": status_code,
            "headers": headers or {},
            "isBase64Encoded": False,
            "body": json.dumps(body),
        }

    def as_response(self) -> dict:
        return self.response_data


class ApiHandler(Handler):
    event_class = Request

    def __init__(
        self,
        func: Callable,
        path: str,
        methods: Iterable[str],
    ) -> None:
        self.path: str = path
        self.methods: Iterable[str] = methods
        super().__init__(func)

    @staticmethod
    def _name_suffix() -> str:
        return "_api"


def route(path: str, methods: Iterable[str] = ("ANY")):
    "Wrapper for creating :class:`ApiHandler` resource."

    def wraper(func):
        return ApiHandler(func, path, methods)

    return wraper


# ___________________ Job ______________________________
# Lambda with EventBridge
# https://docs.aws.amazon.com/lambda/latest/dg/services-cloudwatchevents.html


class JobEvent(LambdaInput):
    @property
    def detail(self) -> dict[str, Any]:
        return self.event["detail"]


class JobHandler(Handler):
    event_class = JobEvent

    def __init__(self, func: Callable, schedule: str) -> None:
        self.schedule = schedule
        super().__init__(func)

    @staticmethod
    def _name_suffix() -> str:
        return "_job"


def job(schedule: str):
    """
    Wrapper for creating :class:`JobHandler` resource.
    Schedule expression [docs](https://docs.aws.amazon.com/eventbridge/latest/userguide/eb-create-rule-schedule.html)
    """

    def wrapper(func):
        return JobHandler(func, schedule)

    return wrapper


# __________________ Worker ____________________________
# Lambda with SQS
# https://docs.aws.amazon.com/lambda/latest/dg/with-sqs.html


class QueueEvent:
    def __init__(self, event: dict) -> None:
        self.event = event
        self._body = None

    def delete(self):
        pass
        # TODO: delete event from queue

    @property
    def message_id(self) -> str:
        return self.event["messageId"]

    @property
    def body(self) -> JsonData:
        if self._body is None:
            try:
                self._body = json.loads(self.event["body"])
            except json.JSONDecodeError:
                self._body = self.event["body"]
        return self._body

    @property
    def attributes(self) -> dict[str, Any]:
        return self.event["attributes"]

    @property
    def message_attributes(self) -> dict[str, Any]:
        return self.event["messageAttributes"]


class SqsEventsSequence(LambdaInput, UserList[QueueEvent]):
    def __init__(self, event: dict, context: dict) -> None:
        super().__init__(event, context)
        self.data = [QueueEvent(e) for e in self.event["Records"]]


class SqsFailedEvents:
    # Returns from lambda failed events
    # https://docs.aws.amazon.com/lambda/latest/dg/with-sqs.html#services-sqs-batchfailurereporting

    def __init__(self, *failed_ids: tuple[str]) -> None:
        self.failed_event_ids: set[str] = set(failed_ids)

    def fail(self, *failed_ids: tuple[str]):
        self.failed_event_ids.update(failed_ids)

    def as_response(self) -> dict:
        return {
            "batchItemFailures": [
                {"itemIdentifier": id for id in self.failed_event_ids}
            ]
        }


class SqsHandler(Handler):
    event_class = SqsEventsSequence

    @staticmethod
    def _name_suffix() -> str:
        return "_worker"

    def __init__(self, func: Callable, queue_name: str) -> None:
        self.queue_name = queue_name
        super().__init__(func)


def sqs_handler(queue_name: str):
    """
    Wrapper for creating :class:`JobHandler` resource.
    """

    def wrapper(func):
        return SqsHandler(
            func,
            queue_name,
        )

    return wrapper



import enum
import os

import boto3

from .base import Handler, ResourceConnector

# ______________________ SQS _______________________________


class SqsPermission(enum.Enum):
    read = 1
    write = 2
    full_access = 3


class SqsConnector(ResourceConnector):
    def __init__(
        self, handler: Handler, resource_name: str, permission: SqsPermission
    ) -> None:
        super().__init__(handler, resource_name)
        self.permission = permission

    def get_resource_client(self):
        if not self._client:
            queue_url = os.environ[f"{self.resource_name.upper()}_URL"]
            sqs = boto3.resource("sqs")
            self._client = sqs.Queue(queue_url)
        return self._client


def sqs(
    queue_name: str,
    permission: SqsPermission = SqsPermission.read,
    # *,
    # attr_name: str = None,
):
    "Add queue `Sqs` resource for :class:`Handler`"

    def wraper(handler: Handler):
        # FIXME: rework how we pass resources into handler
        # handler.extra_kwargs[attr_name or queue_name] = get_queue(queue_name)
        SqsConnector(handler, queue_name, permission)
        return handler

    return wraper



from .base import Resource


class Sqs(Resource):
    def __init__(self, name: str, visibility_timeout: int = 360) -> None:
        super().__init__(name)
        self.visibility_timeout = visibility_timeout



import inspect
import logging
import os
import shutil
import sys
from abc import ABC, abstractmethod
from pathlib import Path
from typing import Generic, TypeVar

from aws_cdk import (
    Duration,
    Stack,
    aws_apigateway,
    aws_events,
    aws_events_targets,
    aws_lambda,
    aws_lambda_event_sources,
    aws_sqs,
)
from constructs import Construct

from viburnum.application import (
    Application,
    Handler,
    Resource,
    ResourceConnector,
    Sqs,
    SqsPermission,
)
from viburnum.application.connectors import SqsConnector
from viburnum.application.handlers import ApiHandler, JobHandler, SqsHandler


def get_builder_class(primitive):
    return getattr(sys.modules[__name__], f"{primitive.__class__.__name__}Builder")


class AppConstruct(Stack):
    def __init__(self, scope: Construct, app: Application, **kwargs) -> None:
        super().__init__(scope, app.name, **kwargs)

        self._app = app
        self._built_resources = {}

        self._prepare_shared_layer()
        self._prepare_libraries_layer()
        self._build_shared_layer()

        self._build_resources()
        self._build_handlers()

    def _build_resources(self):
        for resource in self._app.resources.values():
            builder_class = getattr(
                sys.modules[__name__], f"{resource.__class__.__name__}Builder"
            )
            # FIXME: raise error if there were resources with identical names
            self._built_resources[resource.name] = builder_class(self, resource).build()

    def _build_handlers(self):
        for handler in self._app.handlers:
            handler_class = getattr(
                sys.modules[__name__], f"{handler.__class__.__name__}Builder"
            )
            handler_class(self, handler).build()

    def _prepare_shared_layer(self):
        logging.info("Preparing shared layer")
        shared_layer_folder = Path("./.layers/shared")
        if os.path.exists(shared_layer_folder):
            shutil.rmtree(shared_layer_folder)
        os.mkdir(shared_layer_folder)
        shutil.copytree("shared", str(shared_layer_folder.joinpath("python/shared")))

    def _prepare_libraries_layer(self):
        logging.info("Preparing libraries layer")
        lib_layer_folder = Path("./.layers/lib")
        if os.path.exists(lib_layer_folder):
            shutil.rmtree(lib_layer_folder)
        os.mkdir(lib_layer_folder)
        shutil.copy("requirements.txt", str(lib_layer_folder))

    def _build_shared_layer(self):
        self._shared_layer = aws_lambda.LayerVersion(
            self,
            "SharedLayer",
            code=aws_lambda.Code.from_asset(
                str(Path("./.shared-layer")),
            ),
            compatible_runtimes=[aws_lambda.Runtime.PYTHON_3_9],
        )

    def _build_sqs(self, sqs: Sqs):
        queue = aws_sqs.Queue(
            self, sqs.name, visibility_timeout=Duration.seconds(sqs.visibility_timeout)
        )
        self._built_resources[sqs.name] = queue
        return queue

    def get_built_resource(self, name: str):
        return self._built_resources[name]


# ______________ Handler Builders __________________ #


HandlerType = TypeVar("HandlerType", bound=Handler)


class HandlerBuilder(Generic[HandlerType]):
    def __init__(self, context: "AppConstruct", handler: HandlerType) -> None:
        self.context = context
        self.handler = handler

    def _build_lambda(self):
        dir_ = Path(inspect.getfile(self.handler.func)).parent

        # TODO: add more config options
        lambda_fn = aws_lambda.Function(
            self.context,
            self.handler.name,
            runtime=aws_lambda.Runtime.PYTHON_3_9,
            handler=f"handler.{self.handler.func.__name__}",
            code=aws_lambda.Code.from_asset(str(dir_)),
            environment={
                "APP_NAME": self.context._app.name,
                # "AWS_REGION": self.context.region, This variable is reserved
            },
            layers=[self.context._shared_layer],
        )
        return lambda_fn

    def _connect_resources(self, lambda_: aws_lambda.Function):
        # FIXME: very bad pattern
        for connector in self.handler.resources:
            get_builder_class(connector)(self.context, connector, lambda_).build()

    def build(self):
        # TODO: rework inheritance
        lambda_ = self._build_lambda()
        self._connect_resources(lambda_)
        return lambda_


class ApiHandlerBuilder(HandlerBuilder[ApiHandler]):
    _api = None

    @property
    def api(self) -> aws_apigateway.RestApi:
        if not self.__class__._api:
            self.__class__._api = aws_apigateway.RestApi(
                self.context, f"{self.context._app.name}Api"
            )
        return self.__class__._api

    def build(self):
        lambda_ = super().build()
        self._build_endpoint(lambda_)
        return lambda_

    def _build_endpoint(self, lambda_: aws_lambda.Function):
        integration = aws_apigateway.LambdaIntegration(lambda_)

        endpoint = self.api.root
        for path_part in (p for p in self.handler.path.split("/") if p):
            endpoint = endpoint.get_resource(path_part) or endpoint.add_resource(
                path_part
            )
        for method in self.handler.methods:
            endpoint.add_method(method, integration)


class JobHandlerBuilder(HandlerBuilder[JobHandler]):
    def build(self):
        lambda_ = super().build()
        self._build_rule(lambda_)
        return lambda_

    def _build_rule(self, lambda_):
        aws_events.Rule(
            self.context,
            f"{self.handler.name}_rule",
            schedule=aws_events.Schedule.expression(self.handler.schedule),
            targets=[aws_events_targets.LambdaFunction(lambda_)],
        )


class SqsHandlerBuilder(HandlerBuilder[SqsHandler]):
    def build(self):
        lambda_ = super().build()
        self._handler_connect_queue(lambda_)
        return lambda_

    def _handler_connect_queue(self, lambda_: aws_lambda.Function):
        queue: aws_sqs.Queue = self.context.get_built_resource(self.handler.queue_name)
        # TODO: add more config options including enabling report_batch_item_failure
        _sqs_event_source = aws_lambda_event_sources.SqsEventSource(queue)
        lambda_.add_event_source(_sqs_event_source)


# ______________ Resource Builders __________________ #


ResourceType = TypeVar("ResourceType", bound=Resource)


class ResourceBuilder(ABC, Generic[ResourceType]):
    def __init__(self, context: "AppConstruct", resource: ResourceType) -> None:
        self.context = context
        self.resource = resource

    @abstractmethod
    def build(self):
        ...


class SqsBuilder(ResourceBuilder[Sqs]):
    def build(self):
        # TODO: add more config options
        queue = aws_sqs.Queue(
            self.context,
            self.resource.name,
            visibility_timeout=Duration.seconds(self.resource.visibility_timeout),
        )
        return queue


# _____________________ Resource Conector Builder ________________________

ConnectorType = TypeVar("ConnectorType", bound=ResourceConnector)


class ResourceConnectorBuilder(ABC, Generic[ConnectorType]):
    def __init__(
        self,
        context: "AppConstruct",
        connector: ConnectorType,
        lambda_: aws_lambda.Function,
    ) -> None:
        self.context = context
        self.connector = connector
        self.lambda_ = lambda_

    @abstractmethod
    def build(self):
        ...


class SqsConnectorBuilder(ResourceConnectorBuilder[SqsConnector]):
    def build(self):
        # TODO: validate resource type
        resource: aws_sqs.Queue = self.context.get_built_resource(
            self.connector.resource_name
        )
        if self.connector.permission in [SqsPermission.read, SqsPermission.full_access]:
            resource.grant_consume_messages(self.lambda_)
        if self.connector.permission in [
            SqsPermission.write,
            SqsPermission.full_access,
        ]:
            resource.grant_send_messages(self.lambda_)
        self.lambda_.add_environment(
            f"{self.connector.resource_name.upper()}_URL", resource.queue_url
        )



import weakref
from typing import Callable

# __________________ Resource Connector ___________________


class ResourceConnector:
    def __init__(
        self,
        handler: "Handler",
        resource_name: str,
    ) -> None:
        handler.resources.add(self)
        self.handler = weakref.proxy(handler)
        self.resource_name = resource_name
        self._client = None

    def get_resource_client(self):
        """Return resource client or None"""
        return None


# ___________________ Handler ____________________________


class LambdaInput:
    def __init__(self, event: dict, context: dict) -> None:
        self.event = event
        # TODO: context is not dict type
        self.context = context


class LambdaOutput:
    def as_response(self) -> dict:
        return {}


class Handler:
    event_class = LambdaInput

    def __init__(
        self,
        func: Callable,
    ) -> None:
        self.name = f"{func.__name__}{self._name_suffix()}"
        self.func = func
        self.resources: set[ResourceConnector] = set()
        self.extra_kwargs: dict = {}  # DEPRECATED: useless

    def __call__(self, event: dict, context: dict) -> dict:

        response = self.func(
            self.event_class(event, context), **self._get_resource_clients()
        )
        if isinstance(response, LambdaOutput):
            return response.as_response()
        return response

    def _get_resource_clients(self) -> dict:
        return {r.resource_name: r.get_resource_client() for r in self.resources}

    @staticmethod
    def _name_suffix() -> str:
        """
        Return suffix that will be added to function name
        for creating name for lambda.
        """
        return ""


# __________________________ Resource _____________________________


class Resource:
    def __init__(self, name: str) -> None:
        self.name = name


# __________________________ Application ___________________________


class Application:
    def __init__(self, name: str) -> None:
        self.name: str = name
        self.handlers: list[Handler] = []
        self.resources: dict[str, Resource] = {}

    def add_handler(self, handler: Handler) -> None:
        self.handlers.append(handler)

    def add_resource(self, resource: Resource) -> None:
        self.resources[resource.name] = resource



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['viburnum', 'viburnum.application', 'viburnum.cli', 'viburnum.deployer']

package_data = \
{'': ['*']}

extras_require = \
{'cdk': ['aws-cdk-lib==2.50.0'],
 'constructs': ['constructs>=10.0.0,<11.0.0'],
 'deployer': ['constructs>=10.0.0,<11.0.0', 'aws-cdk-lib==2.50.0']}

setup_kwargs = {
    'name': 'viburnum',
    'version': '0.1.1',
    'description': "It's abstraction on top of AWS CDK, that helps in building serverless web applications.",
    'long_description': "# Viburnum\n\n**Viburum** - it's a small framework built on top of AWS CDK to simplify development and deploying AWS Serverless web applications.\n",
    'author': 'Yaroslav Martynenko',
    'author_email': 'stikblacklabel@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'extras_require': extras_require,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
